from typing import Optional, List
import skimage.transform
import numpy as np
import gym

# IMPORT ROS
import rospy
from std_msgs.msg import UInt64, Float32MultiArray, Bool, Float32
from sensor_msgs.msg import Image

# IMPORT EAGERX
import eagerx.core.register as register
from eagerx.utils.utils import Msg
from eagerx.core.entities import EngineNode
from eagerx.core.constants import process


class ObservationSensor(EngineNode):
    @staticmethod
    @register.spec("ObservationSensor", EngineNode)
    def spec(
        spec,
        name: str,
        rate: float,
        process: Optional[int] = process.BRIDGE,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        color: Optional[str] = "cyan",
    ):
        """ObservationSensor spec"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        spec.initialize(ObservationSensor)

        # Set default
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if inputs else ["tick"]
        spec.config.outputs = outputs if outputs else ["observation"]

    def initialize(self):
        # We will probably use self.simulator[self.obj_name] in callback & reset.
        assert (
            self.process == process.BRIDGE
        ), "Simulation node requires a reference to the simulator, hence it must be launched in the Bridge process"
        self.id = self.bridge_config["env_id"]
        self.obj_name = self.config["name"]
        self.last_obs = None

    @register.states()
    def reset(self):
        self.last_obs = None

    @register.inputs(tick=UInt64)
    @register.outputs(observation=Float32MultiArray)
    def callback(self, t_n: float, tick: Optional[Msg] = None):
        assert isinstance(self.simulator[self.obj_name], dict), (
            'Simulator object "%s" is not compatible with this engine node.' % self.simulator[self.obj_name]
        )
        obs = self.simulator[self.obj_name]["buffer_obs"]
        self.simulator[self.obj_name]["buffer_obs"] = []
        if len(obs) == 0:
            obs = self.last_obs
        else:
            obs = obs[-1]
            self.last_obs = obs
        return dict(observation=Float32MultiArray(data=obs))


class RewardSensor(EngineNode):
    @staticmethod
    @register.spec("RewardSensor", EngineNode)
    def spec(
        spec,
        name: str,
        rate: float,
        process: Optional[int] = process.BRIDGE,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        color: Optional[str] = "cyan",
    ):
        """RewardSensor spec"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        spec.initialize(RewardSensor)

        # Set default
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if inputs else ["tick"]
        spec.config.outputs = outputs if outputs else ["reward"]

    def initialize(self):
        # We will probably use self.simulator[self.obj_name] in callback & reset.
        assert (
            self.process == process.BRIDGE
        ), "Simulation node requires a reference to the simulator, hence it must be launched in the Bridge process"
        self.id = self.bridge_config["env_id"]
        self.obj_name = self.config["name"]
        self.last_reward = None

    @register.states()
    def reset(self):
        self.last_reward = 0.0

    @register.inputs(tick=UInt64)
    @register.outputs(reward=Float32)
    def callback(self, t_n: float, tick: Optional[Msg] = None):
        assert isinstance(self.simulator[self.obj_name], dict), (
            'Simulator object "%s" is not compatible with this engine node.' % self.simulator[self.obj_name]
        )
        reward = self.simulator[self.obj_name]["buffer_reward"]
        self.simulator[self.obj_name]["buffer_reward"] = []
        if len(reward) == 0:
            reward = self.last_reward
        else:
            self.last_reward = reward[-1]
            reward = sum(reward)
        return dict(reward=Float32(data=reward))


class DoneSensor(EngineNode):
    @staticmethod
    @register.spec("DoneSensor", EngineNode)
    def spec(
        spec,
        name: str,
        rate: float,
        process: Optional[int] = process.BRIDGE,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        color: Optional[str] = "cyan",
    ):
        """DoneSensor spec"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        spec.initialize(DoneSensor)

        # Set default
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if inputs else ["tick"]
        spec.config.outputs = outputs if outputs else ["done"]

    def initialize(self):
        # We will probably use self.simulator[self.obj_name] in callback & reset.
        assert (
            self.process == process.BRIDGE
        ), "Simulation node requires a reference to the simulator, hence it must be launched in the Bridge process"
        self.id = self.bridge_config["env_id"]
        self.obj_name = self.config["name"]
        self.last_done = None

    @register.states()
    def reset(self):
        self.last_done = False

    @register.inputs(tick=UInt64)
    @register.outputs(done=Bool)
    def callback(self, t_n: float, tick: Optional[Msg] = None):
        assert isinstance(self.simulator[self.obj_name], dict), (
            'Simulator object "%s" is not compatible with this engine node.' % self.simulator[self.obj_name]
        )
        done = self.simulator[self.obj_name]["buffer_done"]
        self.simulator[self.obj_name]["buffer_done"] = []
        if len(done) == 0:
            done = self.last_done
        else:
            done = any(done) or self.last_done
            self.last_done = done
        return dict(done=Bool(data=done))


class ActionActuator(EngineNode):
    @staticmethod
    @register.spec("ActionActuator", EngineNode)
    def spec(
        spec,
        name: str,
        rate: float,
        zero_action=None,
        process: Optional[int] = process.BRIDGE,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        color: Optional[str] = "green",
    ):
        """ActionActuator spec"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        spec.initialize(ActionActuator)

        # Set default
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if inputs else ["tick", "action"]
        spec.config.outputs = outputs if outputs else ["action_applied"]

        # Modify custom node params
        spec.config.zero_action = zero_action

    def initialize(self, zero_action):
        # We will probably use self.simulator[self.obj_name] in callback & reset.
        assert (
            self.process == process.BRIDGE
        ), "Simulation node requires a reference to the simulator, hence it must be launched in the Bridge process"
        self.obj_name = self.config["name"]
        self.simulator[self.obj_name]["env"]: gym.Env
        self.is_discrete = (
            True if isinstance(self.simulator[self.obj_name]["env"].action_space, gym.spaces.Discrete) else False
        )
        if zero_action is None:
            self.zero_action = self.simulator[self.obj_name]["env"].action_space.sample()
        else:
            if isinstance(zero_action, list):
                dtype = self.simulator[self.obj_name]["env"].action_space.dtype
                self.zero_action = np.array(zero_action, dtype=dtype)
            else:
                self.zero_action = zero_action
            assert self.simulator[self.obj_name]["env"].action_space.contains(
                self.zero_action
            ), f'The zero action provided for "{self.obj_name}" is not contained in the action space of this environment.'

    @register.states()
    def reset(self):
        # This controller is stateless (in contrast to e.g. a PID controller).
        self.simulator[self.obj_name]["next_action"] = self.zero_action

    @register.inputs(tick=UInt64, action=Float32MultiArray)
    @register.outputs(action_applied=Float32MultiArray)
    def callback(
        self,
        t_n: float,
        tick: Optional[Msg] = None,
        action: Optional[Msg] = None,
    ):
        assert isinstance(self.simulator[self.obj_name], dict), (
            'Simulator object "%s" is not compatible with this engine node.' % self.simulator[self.obj_name]
        )

        # Set action in simulator for next step.
        if len(action.msgs) > 0:
            self.simulator[self.obj_name]["next_action"] = (
                int(action.msgs[-1].data[0]) if self.is_discrete else action.msgs[-1].data
            )
        else:
            self.simulator[self.obj_name]["next_action"] = self.zero_action

        # Prepare output message
        action_applied = self.simulator[self.obj_name]["next_action"]

        # Send action that has been applied.
        return dict(action_applied=Float32MultiArray(data=action_applied))


class GymImage(EngineNode):
    @staticmethod
    @register.spec("GymImage", EngineNode)
    def spec(
        spec,
        name: str,
        rate: float,
        process: Optional[int] = process.BRIDGE,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        color: Optional[str] = "cyan",
        shape: Optional[List[int]] = None,
        always_render=False,
    ):
        """GymImage spec"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        spec.initialize(GymImage)

        # Set default
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if inputs else ["tick"]
        spec.config.outputs = outputs if outputs else ["image"]

        # Modify custom node params
        spec.config.shape = shape if shape else [200, 200]
        spec.config.always_render = always_render

    def initialize(self, shape, always_render):
        # We will probably use self.simulator[self.obj_name] in callback & reset.
        assert (
            self.process == process.BRIDGE
        ), "Simulation node requires a reference to the simulator, hence it must be launched in the Bridge process"
        self.shape = tuple(shape)
        self.always_render = always_render
        self.render_toggle = False
        self.id = self.bridge_config["env_id"]
        self.obj_name = self.config["name"]
        self.render_toggle_pub = rospy.Subscriber("%s/env/render/toggle" % self.ns, Bool, self._set_render_toggle)

    @register.states()
    def reset(self):
        # This sensor is stateless (in contrast to e.g. a Kalman filter).
        pass

    @register.inputs(tick=UInt64)
    @register.outputs(image=Image)
    def callback(self, t_n: float, tick: Optional[Msg] = None):
        assert isinstance(self.simulator[self.obj_name], dict), (
            'Simulator object "%s" is not compatible with this engine node.' % self.simulator[self.obj_name]
        )
        if self.always_render or self.render_toggle:
            rgb = self.simulator[self.obj_name]["env"].render(mode="rgb_array")

            # Resize image if not matching desired self.shape (defined in .yaml)
            if rgb.shape[:2] != tuple(self.shape):
                kwargs = dict(output_shape=self.shape, mode="edge", order=1, preserve_range=True)
                rgb = skimage.transform.resize(rgb, **kwargs).astype(rgb.dtype)

            # Prepare ROS msg
            height = rgb.shape[0]
            width = rgb.shape[1]
            data = rgb.tobytes("C")
            msg = Image(data=data, height=height, width=width, encoding="rgb8")
            # self._show_ros_image(msg)
        else:
            msg = Image()
        return dict(image=msg)

    def _show_ros_image(self, msg):
        import matplotlib.pyplot as plt

        rgb_back = np.frombuffer(msg.data, dtype=np.uint8).reshape(msg.height, msg.width, -1)
        plt.imshow(rgb_back)
        plt.show()

    def _set_render_toggle(self, msg):
        if msg.data:
            rospy.loginfo("[%s] START RENDERING!" % self.name)
        else:
            self.simulator[self.obj_name]["env"].close()
            rospy.loginfo("[%s] STOPPED RENDERING!" % self.name)
        self.render_toggle = msg.data



from eagerx.core.entities import BaseConverter
import eagerx.core.register as register


class Identity(BaseConverter):
    """The Identity converter that mimics the API of any other converter but simply passes through all messages."""

    def __init__(self):
        super().__init__()

    @staticmethod
    @register.spec("Identity", BaseConverter)
    def spec(spec):
        pass

    @staticmethod
    def get_opposite_msg_type(cls, msg_type):
        return msg_type

    def convert(self, msg):
        return msg

    def initialize(self):
        pass



# OTHER IMPORTS
from typing import Optional, List
from math import isclose

# ROS IMPORTS
import rospy
from std_msgs.msg import UInt64, String, Bool

# EAGERx IMPORTS
from eagerx.core.constants import process
from eagerx.utils.utils import Msg
from eagerx.core.entities import Node, ResetNode, EngineNode, SpaceConverter
import eagerx.core.register as register


class RealResetNode(ResetNode):
    def initialize(self, test_arg, test_kwarg="test"):
        pass

    @staticmethod
    @register.spec("RealReset", ResetNode)
    def spec(
        spec,
        name: str,
        rate: float,
        process: Optional[int] = process.ENVIRONMENT,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        states: Optional[List[str]] = None,
        targets: Optional[List[str]] = None,
        color: Optional[str] = "blue",
        test_arg: Optional[str] = "test_argument",
    ):
        """Create the default spec that is compatible with __init__(...), .callback(...), and .reset(...)"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        spec.initialize(RealResetNode)

        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if inputs else ["in_1"]
        spec.config.outputs = outputs if outputs else ["out_1"]
        spec.config.states = states if states else ["state_1"]
        spec.config.targets = targets if targets else ["target_1"]

        # Modify extra params
        spec.config.test_arg = test_arg

        # Test NodeSPec
        spec.add_target("target_test", UInt64)
        spec.remove_target("target_test")

        # set input parameters
        spec.inputs.in_1.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.inputs.in_2.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.inputs.in_1.window = 0
        spec.inputs.in_2.window = 0

        # Set outputs parameters
        spec.outputs.out_1.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.outputs.out_2.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")

        # Set states parameters
        spec.states.state_1.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        return spec

    @register.states(state_1=UInt64)
    def reset(self, state_1: Optional[UInt64] = None) -> None:
        if "in_1" in [i["name"] for i in self.inputs]:
            self.set_delay(0.0, "inputs", "in_1")

    @register.inputs(in_1=UInt64, in_2=UInt64)
    @register.outputs(out_1=UInt64, out_2=UInt64)
    @register.targets(target_1=UInt64)
    def callback(
        self,
        t_n: float,
        in_1: Optional[Msg] = None,
        in_2: Optional[Msg] = None,
        target_1: Optional[Msg] = None,
    ):
        # output type is always Dict[str, Union[UInt64, output_msg_types]] because done flags are also inside the output_msgs
        inputs = {"in_1": in_1, "in_2": in_2}
        pop_keys = []
        for key, value in inputs.items():
            if value is None:
                pop_keys.append(key)
        [inputs.pop(i) for i in pop_keys]

        # Verify that all timestamps are smaller or equal to node time
        for i in self.inputs:
            name = i["name"]
            if name in inputs:
                t_i = inputs[name].info.t_in
                if len(t_i) > 0 and not all((t.sim_stamp - t_n) <= 1e-7 for t in t_i if t is not None):
                    rospy.logerr(f"[{self.name}][{name}]: Not all t_i are smaller or equal to t_n.")

        # Fill output msg with number of node ticks
        output_msgs = dict()
        Nc = self.num_ticks
        for i in self.outputs:
            name = i["name"]
            output_msgs[name] = UInt64(data=Nc)

        # Fill state done msg with number of node ticks
        for i in self.targets:
            name = i["name"]
            msg = Bool()
            if self.num_ticks > 5:
                msg.data = True
            else:
                msg.data = False
            output_msgs[name + "/done"] = msg
        return output_msgs


class TestNode(EngineNode):
    def initialize(self, test_arg):
        pass

    @register.states(state_1=UInt64, state_2=UInt64)
    def reset(self, state_1: Optional[UInt64] = None, state_2: Optional[UInt64] = None) -> None:
        return

    @register.inputs(in_1=UInt64, in_2=UInt64, in_3=String, tick=UInt64)
    @register.outputs(out_1=UInt64, out_2=UInt64)
    def callback(
        self,
        t_n: float,
        in_1: Optional[Msg] = None,
        in_2: Optional[Msg] = None,
        in_3: Optional[Msg] = None,
        tick: Optional[Msg] = None,
    ):
        inputs = {"in_1": in_1, "in_2": in_2, "tick": tick}
        pop_keys = []
        for key, value in inputs.items():
            if value is None:
                pop_keys.append(key)
        [inputs.pop(i) for i in pop_keys]

        # Verify that # of ticks equals internal counter
        node_tick = t_n * self.rate
        if self.is_reactive and not isclose(self.num_ticks, node_tick):
            rospy.logerr(
                f"[{self.name}][callback]: ticks not equal (self.num_ticks={self.num_ticks}, node_tick={round(node_tick)})."
            )
            pass

        # Verify that all timestamps are smaller or equal to node time
        t_n = node_tick * (1 / self.rate)
        for i in self.inputs:
            name = i["name"]
            if name in inputs:
                t_i = inputs[name].info.t_in
                if len(t_i) > 0 and not all((t.sim_stamp - t_n) <= 1e-7 for t in t_i if t is not None):
                    rospy.logerr(f"[{self.name}][{name}]: Not all t_i are smaller or equal to t_n.")

        # Fill output msg with number of node ticks
        output_msgs = dict()
        Nc = self.num_ticks
        for i in self.outputs:
            name = i["name"]
            output_msgs[name] = UInt64(data=Nc)
        return output_msgs


class ProcessNode(TestNode):
    @staticmethod
    @register.spec("Process", Node)
    def spec(
        spec,
        name: str,
        rate: float,
        process: Optional[int] = process.ENVIRONMENT,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        states: Optional[List[str]] = None,
        color: Optional[str] = "white",
        test_arg: Optional[str] = "test_argument",
    ):
        """ProcessNode spec"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        spec.initialize(ProcessNode)

        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if inputs else ["in_1"]
        spec.config.outputs = outputs if outputs else ["out_1"]
        spec.config.states = states if states else ["state_1"]

        # Modify extra params
        spec.config.test_arg = test_arg

        # set input parameters
        spec.inputs.in_1.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.inputs.in_2.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.inputs.in_3.space_converter = SpaceConverter.make("Space_RosString", [0], [100], dtype="uint64")
        spec.inputs.in_1.window = 0
        spec.inputs.in_2.window = 0
        spec.inputs.in_3.window = 0
        spec.inputs.tick.window = 0

        # Set outputs parameters
        spec.outputs.out_1.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.outputs.out_2.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")

        # Set states parameters
        spec.states.state_1.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.states.state_2.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        return spec


class KalmanNode(TestNode):
    @staticmethod
    @register.spec("KalmanFilter", Node)
    def spec(
        spec,
        name: str,
        rate: float,
        process: Optional[int] = process.ENVIRONMENT,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        states: Optional[List[str]] = None,
        color: Optional[str] = "grey",
        test_arg: Optional[str] = "test_argument",
    ):
        """KalmanNode spec"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        spec.initialize(KalmanNode)

        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if inputs else ["in_1"]
        spec.config.outputs = outputs if outputs else ["out_1"]
        spec.config.states = states if states else ["state_1"]

        # Modify extra params
        spec.config.test_arg = test_arg

        # Test NodeSpec adding & removing of components
        spec.add_input("in_test", UInt64)
        spec.add_output("out_test", UInt64)
        spec.add_state(
            "state_test", UInt64, space_converter=SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        )
        try:
            spec.add_target("target_test", UInt64)
        except AssertionError as e:
            print(e)
        spec.remove_input("in_test")
        spec.remove_output("out_test")
        spec.remove_state("state_test")

        # Test getting parameters
        _ = spec.inputs.tick.msg_type

        # Remove unused inputs
        spec.remove_input("tick")
        spec.remove_input("in_3")

        # Set input parameters
        spec.inputs.in_1.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.inputs.in_2.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.inputs.in_1.window = 0
        spec.inputs.in_2.window = 0

        # Set outputs parameters
        spec.outputs.out_1.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.outputs.out_2.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")

        # Set states parameters
        spec.states.state_1.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.states.state_2.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        return spec


class TestActuator(TestNode):
    @staticmethod
    @register.spec("TestActuator", EngineNode)
    def spec(
        spec,
        name: str,
        rate: float,
        process: Optional[int] = process.BRIDGE,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        color: Optional[str] = "green",
        test_arg: Optional[str] = "test_argument",
    ):
        """TestActuator spec"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        spec.initialize(TestActuator)

        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if inputs else ["tick", "in_1", "in_2", "in_3"]
        spec.config.outputs = outputs if outputs else ["out_1"]

        # Modify extra params
        spec.config.test_arg = test_arg

        # Set state parameters
        spec.states.state_1.space_converter = SpaceConverter.make("Space_RosUInt64", low=[0], high=[100], dtype="uint64")
        spec.states.state_2.space_converter = SpaceConverter.make("Space_RosUInt64", low=[0], high=[100], dtype="uint64")
        return spec


class TestSensor(TestNode):
    @staticmethod
    @register.spec("TestSensor", EngineNode)
    def spec(
        spec,
        name: str,
        rate: float,
        process: Optional[int] = process.BRIDGE,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        states: Optional[List[str]] = None,
        color: Optional[str] = "cyan",
        test_arg: Optional[str] = "test_argument",
    ):
        """TestSensor spec"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        spec.initialize(TestSensor)

        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = inputs if inputs else ["tick_1", "in_1"]
        spec.config.outputs = outputs if outputs else ["out_1"]
        spec.config.states = states if states else ["state_1"]

        # Modify extra params
        spec.config.test_arg = test_arg

        # Set state parameters
        spec.states.state_1.space_converter = SpaceConverter.make("Space_RosUInt64", low=[0], high=[100], dtype="uint64")
        spec.states.state_2.space_converter = SpaceConverter.make("Space_RosUInt64", low=[0], high=[100], dtype="uint64")
        return spec



from eagerx.core.entities import SpaceConverter, Converter, Processor
import eagerx.core.register as register

# Converter specific
from std_msgs.msg import UInt64, String
from sensor_msgs.msg import Image
import numpy as np
from gym.spaces import Box


class Space_RosUInt64(SpaceConverter):
    MSG_TYPE_A = np.ndarray
    MSG_TYPE_B = UInt64

    def initialize(self, low, high, shape=None, dtype="uint64"):
        self.low = np.array(low)
        self.high = np.array(high)
        self.shape = shape
        self.dtype = dtype

    @staticmethod
    @register.spec("Space_RosUInt64", SpaceConverter)
    def spec(spec, low, high, shape=None, dtype="uint64"):
        # Initialize converter
        Space_RosUInt64.initialize_spec(spec)

        spec.config.low = low
        spec.config.high = high
        spec.config.shape = shape
        spec.config.dtype = dtype

        # Test spec
        spec.config.dtype = spec.config.dtype

    def get_space(self):
        return Box(self.low, self.high, shape=self.shape, dtype=self.dtype)

    def A_to_B(self, msg):
        return UInt64(data=msg[0])

    def B_to_A(self, msg):
        return np.array([msg.data], dtype=self.dtype)


class Space_RosString(SpaceConverter):
    MSG_TYPE_A = np.ndarray
    MSG_TYPE_B = String

    def initialize(self, low, high, shape=None, dtype="uint64"):
        self.low = low
        self.high = high
        if isinstance(low, list):
            self.low = np.array(self.low)
        if isinstance(high, list):
            self.high = np.array(self.high)
        if shape is not None:
            assert (
                isinstance(low, (int, float)) or self.low.shape == shape
            ), "If a shape is defined, low must be of a list or type (int, float)."
            assert (
                isinstance(high, (int, float)) or self.high.shape == shape
            ), "If a shape is defined, high must be a list of type (int, float)."
            self.low = low
            self.high = high
        self.shape = shape
        self.dtype = dtype

    @staticmethod
    @register.spec("Space_RosString", SpaceConverter)
    def spec(spec, low, high, shape=None, dtype="uint64"):
        # Initialize converter
        Space_RosString.initialize_spec(spec)

        spec.config.low = low
        spec.config.high = high
        spec.config.shape = shape
        spec.config.dtype = dtype

    def get_space(self):
        return Box(self.low, self.high, shape=self.shape, dtype=self.dtype)

    def A_to_B(self, msg):
        return String(data="string: %s " % msg[0])

    def B_to_A(self, msg):
        return np.array([int(msg.data[8:])], dtype=self.dtype)


class RosImage_RosUInt64(Converter):
    MSG_TYPE_A = Image
    MSG_TYPE_B = UInt64

    def initialize(self, test_arg):
        self.test_arg = test_arg

    @staticmethod
    @register.spec("RosImage_RosUInt64", Converter)
    def spec(spec, test_arg):
        # Initialize converter
        RosImage_RosUInt64.initialize_spec(spec)

        spec.config.test_arg = test_arg

    def A_to_B(self, msg):
        return UInt64(data=999)

    def B_to_A(self, msg):
        return Image()


class RosString_RosUInt64(Converter):
    MSG_TYPE_A = String
    MSG_TYPE_B = UInt64

    def initialize(self, test_arg):
        self.test_arg = test_arg

    @staticmethod
    @register.spec("RosString_RosUInt64", Converter)
    def spec(spec, test_arg):
        # Initialize converter
        spec.initialize(RosString_RosUInt64)

        spec.config.test_arg = test_arg

    def A_to_B(self, msg):
        return UInt64(data=int(msg.data[8:]))

    def B_to_A(self, msg):
        return String(data="string: %s" % msg.data)


class IdentityProcessor(Processor):
    MSG_TYPE = UInt64

    @staticmethod
    @register.spec("IdentityProcessor", Processor)
    def spec(spec):
        IdentityProcessor.initialize_spec(spec)

    def initialize(self):
        pass

    def convert(self, msg):
        return msg



# ROS IMPORTS
from std_msgs.msg import Float32MultiArray

# RX IMPORTS
import eagerx.core.register as register
from eagerx.core.entities import SpaceConverter

# OTHER
import numpy as np
from gym.spaces import Box
import gym


class GymSpace_Float32MultiArray(SpaceConverter):
    MSG_TYPE_A = np.ndarray
    MSG_TYPE_B = Float32MultiArray

    @staticmethod
    @register.spec("GymSpace_Float32MultiArray", SpaceConverter)
    def spec(
        spec,
        gym_id: str = None,
        space: str = None,
        low=None,
        high=None,
        shape=None,
        dtype="float32",
    ):
        # Initialize converter
        spec.initialize(GymSpace_Float32MultiArray)

        spec.config.gym_id = gym_id
        spec.config.space = space
        spec.config.low = low
        spec.config.high = high
        spec.config.shape = shape
        spec.config.dtype = dtype

    def initialize(self, gym_id=None, space=None, low=None, high=None, shape=None, dtype="float32"):
        if gym_id is not None and space in ["observation", "action"]:
            self.env = gym.make(gym_id)
            if space == "observation":
                self.space = self.env.observation_space
            else:  # space == 'action
                self.space = self.env.action_space
            if isinstance(self.space, gym.spaces.Discrete):
                self.space = gym.spaces.Box(low=0, high=self.space.n - 1, shape=(1,), dtype="int64")
        else:
            self.space = None
            self.low = np.array(low)
            self.high = np.array(high)
            self.shape = shape
        self.dtype = dtype

    def get_space(self):
        if self.space is not None:
            return self.space
        else:
            return Box(self.low, self.high, shape=self.shape, dtype=self.dtype)

    def A_to_B(self, msg):
        return Float32MultiArray(data=msg)

    def B_to_A(self, msg):
        return np.array(msg.data, dtype=self.dtype)



import functools
import inspect
import rospy
import copy
from eagerx.utils.utils import deepcopy


# Global registry with registered entities (bridges, objects, nodes, converters, simnodes, etc..)
REGISTRY = dict()
# Global registry with registered I/O types of (.callback, .reset/.pre_reset, .initialize, .add_object)
TYPE_REGISTER = dict()


class ReverseRegisterLookup:
    def __init__(self, d):
        self._dict = d

    def __getitem__(self, spec_lookup):
        for _entity_cls, entity_id in self._dict.items():
            for entity_id, entry in entity_id.items():
                if entry["spec"] == spec_lookup:
                    return entity_id


class LookupType:
    def __init__(self, d):
        self._dict = d

    @deepcopy
    def __getitem__(self, func_lookup):
        name_split = func_lookup.__qualname__.split(".")
        cls_name = name_split[0]
        # fn_name = name_split[1]
        return self._dict[cls_name]


# Global (reversed) registry of REGISTER:
REVERSE_REGISTRY = ReverseRegisterLookup(REGISTRY)
LOOKUP_TYPES = LookupType(TYPE_REGISTER)


def spec(entity_id, entity_cls):
    """Register a spec function to make an entity"""

    def _register(func):
        entry = func.__module__ + "/" + func.__qualname__
        rospy.logdebug("[register]: entity_id=%s, entity=%s, entry=%s" % (entity_id, entity_cls.__name__, entry))

        @functools.wraps(functools.partial(func, None))
        def _spec(*args, **kwargs):
            """Make an entity with the registered spec function"""
            entity_type = func.__module__ + "/" + func.__qualname__[:-5]
            if not entity_id == "Identity":
                rospy.logdebug("[make]: entity_id=%s, entity=%s, entry=%s" % (entity_id, entity_cls.__name__, entity_type))
            spec = entity_cls.pre_make(entity_id, entity_type)
            try:
                func(spec, *args, **kwargs)
            except TypeError as e:
                if "spec()" in e.args[0]:
                    signature = entity_cls.get_spec(entity_id, verbose=False)
                    err = (
                        f'You can only specify arguments according to the signature of the spec function of "{entity_id}".\n\n'
                    )
                    err += f"The signature for this spec looks like: \n\n{signature}"
                    raise TypeError(err) from e
                else:
                    raise
            return spec

        if entity_cls not in REGISTRY:
            """Add entity if this is the first registration of entity kind"""
            REGISTRY[entity_cls] = dict()
        if entity_id in REGISTRY[entity_cls]:
            """Check if spec of duplicate entity_id corresponds to same spec function"""
            assert (
                _spec == REGISTRY[entity_cls][entity_id]["spec"]
            ), f'There is already a {entity_cls.__name__} with entity_id "{entity_id}" registered.'
        cls = f"{func.__module__}/{func.__qualname__[:-5]}"
        REGISTRY[entity_cls][entity_id] = {"spec": _spec, "cls": cls}
        return _spec

    return _register


def make(entity, id, *args, **kwargs):
    """Make an entity with the registered spec function"""
    assert entity in REGISTRY, f'No entities of type "{entity.__name__}" registered.'
    assert (
        id in REGISTRY[entity]
    ), f'No entities of type "{entity.__name__}" registered under id "{id}". Available entities "{[s for s in list(REGISTRY[entity].keys())]}".'
    return REGISTRY[entity][id]["spec"](*args, **kwargs)


def get_spec(entity, id, verbose=True):
    """Get information on the entity's spec function"""
    if verbose:
        help(REGISTRY[entity][id]["spec"])
    return inspect.signature(REGISTRY[entity][id]["spec"])


# TYPES


def _register_types(TYPE_REGISTER, component, cnames, func, cls_only=True):
    name_split = func.__qualname__.split(".")
    cls_name = name_split[0]
    fn_name = name_split[1]
    entry = func.__module__ + "/" + func.__qualname__
    if cls_only:
        for key, cls in cnames.items():
            assert inspect.isclass(
                cls
            ), f'TYPE REGISTRATION ERROR: [{cls_name}][{fn_name}][{component}]: An instance "{cls}" of class "{cls.__class__}" was provided for "{key}". Please provide the class instead.'
    rospy.logdebug(f"[{cls_name}][{fn_name}]: {component}={cnames}, entry={entry}")

    @functools.wraps(func)
    def registered_fn(*args, **kwargs):
        """Call the registered function"""
        # todo: perform type checking in I/O using TYPE_REGISTER
        return func(*args, **kwargs)

    if cls_name not in TYPE_REGISTER:
        """Add class if this is the first registration of class kind"""
        TYPE_REGISTER[cls_name] = dict()

    if component in TYPE_REGISTER[cls_name]:
        """Check if already registered component of duplicate function matches."""
        rospy.logdebug(f"[{cls_name}][{component}]: {component}={cnames}, entry={entry}")
        assert (
            cnames == TYPE_REGISTER[cls_name][component]
        ), f'There is already a [{cls_name}][{component}] registered with cnames "{TYPE_REGISTER[cls_name][component]}", and they do not match the cnames of this function: "{cnames}".'
    TYPE_REGISTER[cls_name][component] = cnames
    return registered_fn


def inputs(**inputs):
    """Register input msg_types of callback"""
    return functools.partial(_register_types, TYPE_REGISTER, "inputs", inputs)


def outputs(**outputs):
    """Register output msg_types of callback"""
    return functools.partial(_register_types, TYPE_REGISTER, "outputs", outputs)


def states(**states):
    """Register state msg_types of reset"""
    return functools.partial(_register_types, TYPE_REGISTER, "states", states)


def targets(**targets):
    """Register target msg_types of callback"""
    return functools.partial(_register_types, TYPE_REGISTER, "targets", targets)


def sensors(**sensors):
    """Register sensor msg_types"""
    return functools.partial(_register_types, TYPE_REGISTER, "sensors", sensors)


def actuators(**actuators):
    """Register actuator msg_types"""
    return functools.partial(_register_types, TYPE_REGISTER, "actuators", actuators)


def simstates(**simstates):
    """Register simstates msg_type"""
    return functools.partial(_register_types, TYPE_REGISTER, "states", simstates)


def bridge_config(**bridge_config):
    """Register default bridge_config arguments"""
    return functools.partial(_register_types, TYPE_REGISTER, "bridge_config", bridge_config, cls_only=False)


def config(**params):
    """Register default config arguments"""
    return functools.partial(
        _register_types,
        TYPE_REGISTER,
        "config",
        params,
        cls_only=False,
    )


# BRIDGES


def bridge(entity_id, bridge_cls):
    bridge_config = LOOKUP_TYPES[bridge_cls.add_object]["bridge_config"]
    bridge_id = REVERSE_REGISTRY[bridge_cls.spec]

    def _register(func):
        name_split = func.__qualname__.split(".")
        cls_name = name_split[0]
        fn_name = name_split[1]
        entry = func.__module__ + "/" + func.__qualname__
        rospy.logdebug(f"[{cls_name}][{fn_name}]: bridge_id={bridge_id}, entry={entry}")

        @functools.wraps(func)
        def _bridge(spec):
            """First, initialize spec with object_info, then call the bridge function"""
            # Add default bridge_config parameters
            spec._initialize_bridge_config(bridge_id, copy.deepcopy(bridge_config))
            # Initialize engine graph
            graph = spec._initialize_object_graph()
            # Modify bridge_config with user-defined bridge implementation
            func(spec, graph)
            # Add graph to spec & remove redundant states
            spec._add_graph(bridge_id, graph)

        # Register bridge implementation for object
        from eagerx.core.entities import Object

        msg = f"Cannot register bridge '{bridge_id}' for object '{entity_id}'. "
        assert Object in REGISTRY, msg + "No Objects have been registered. Make sure to import the object."
        assert entity_id in REGISTRY[Object], msg + "No object with this id was registered. Make sure to import the object."

        # Check if spec of duplicate entity_id corresponds to same spec function
        flag = bridge_id in REGISTRY[Object][entity_id] and _bridge == REGISTRY[Object][entity_id][bridge_id]
        assert not flag, msg + "There is already a bridge implementation of this type registered."

        # Register bridge implementation
        REGISTRY[Object][entity_id][bridge_id] = _bridge
        return _bridge

    return _register


def add_bridge(spec, bridge_id):
    """Add bridge based on registered id"""
    entity_id = spec.config.entity_id

    # Register bridge implementation for object
    from eagerx.core.entities import Object

    msg = f"Cannot ad bridge implementation '{bridge_id}' for object '{entity_id}'. "
    assert Object in REGISTRY, msg + "No Objects have been registered. Make sure to import the object."
    assert entity_id in REGISTRY[Object], msg + "No object with this id was registered. Make sure to import the object."

    _ = REGISTRY[Object][entity_id][bridge_id](spec)



# ROS SPECIFIC
import rospy
import rosgraph
import roslaunch

# from roslaunch import substitution_args
# from roslaunch.substitution_args import _collect_args

# OTHER
from typing import List, NamedTuple, Any, Optional, Dict, Union, Tuple
import time
import importlib
import inspect
from functools import wraps
from time import sleep
import copy
import ast
import json


def dict_null(items):
    result = {}
    for key, value in items:
        if value is None:
            value = "null"
        result[key] = value
    return result


def dict_None(items):
    result = {}
    for key, value in items:
        if value == "null":
            value = None
        result[key] = value
    return result


def replace_None(d, to_null=True):
    dict_str = json.dumps(d)
    if to_null:
        object_pairs_hook = dict_null
    else:
        object_pairs_hook = dict_None
    return json.loads(dict_str, object_pairs_hook=object_pairs_hook)


def get_attribute_from_module(attribute, module=None):
    if module is None:
        module, attribute = attribute.split("/")
    module = importlib.import_module(module)
    attribute = getattr(module, attribute)
    return attribute


def initialize_converter(args):
    converter_args = copy.deepcopy(args)
    converter_args.pop("converter_type")
    converter_cls = get_attribute_from_module(args["converter_type"])
    return converter_cls(**converter_args)


def initialize_state(args):
    state_cls = get_attribute_from_module(args["state_type"])
    del args["state_type"]
    return state_cls(**args)


def get_opposite_msg_cls(msg_type, converter_cls):
    if isinstance(msg_type, str):
        msg_type = get_attribute_from_module(msg_type)
    if isinstance(converter_cls, dict):
        converter_cls = get_attribute_from_module(converter_cls["converter_type"])
    return converter_cls.get_opposite_msg_type(converter_cls, msg_type)


def get_opposite_msg_cls_v2(msg_type, converter_cls):
    if isinstance(msg_type, str):
        msg_type_cls = get_attribute_from_module(msg_type)
    else:
        msg_type_cls = msg_type
    if isinstance(converter_cls, dict):
        converter_cls = get_attribute_from_module(converter_cls["converter_type"])
    else:  # It is a View object
        converter_cls = converter_cls.to_dict()
        converter_cls = get_attribute_from_module(converter_cls["converter_type"])
    opposite_msg_type = converter_cls.get_opposite_msg_type(converter_cls, msg_type_cls)
    if isinstance(msg_type, str):
        return get_module_type_string(opposite_msg_type)
    else:
        return opposite_msg_type


def get_module_type_string(cls):
    module = inspect.getmodule(cls).__name__
    return "%s/%s" % (module, cls.__name__)


def get_cls_from_string(cls_string):
    return get_attribute_from_module(cls_string)


def get_param_with_blocking(name, timeout=5):
    params = None
    start = time.time()
    it = 0
    while params is None:

        try:
            params = rospy.get_param(name)
        except (rosgraph.masterapi.Error, KeyError):
            sleep_time = 0.01
            if it % 20 == 0:
                rospy.logdebug(
                    'Parameters under namespace "%s" not (yet) uploaded on parameter server. Retry with small pause (%s s).'
                    % (name, sleep_time)
                )
            sleep(sleep_time)
            pass
        if time.time() - start > timeout:
            break
        it += 1

    return replace_None(params, to_null=False)


def substitute_args(
    param: Union[str, Dict],
    context: Optional[Dict] = None,
    only: Optional[List[str]] = None,
):
    """Substitute arguments based on the context dictionairy (and possibly sourced packages).
    Follows the xacro substition convention of ROS, with a 'config' and 'ns' command added.
    :param param: dict or string we wish to perform substitutions on.
    :param context: dict[command][context] with replacement values .
    :param only: List of possible commands. If only is not provided, all commands are executed.
                 Options are ['env', 'optenv', 'dirname', 'anon', 'arg', 'ns', 'config'].
    :return: Substituted param file
    """
    # substitute string
    if isinstance(param, str):
        param = resolve_args(param, context, only=only)
        return param

    # For every key in the dictionary (not performing copy.deepcopy!)
    if isinstance(param, dict):
        for key in param:
            # If the value is of type `(Ordered)dict`, then recurse with the value
            if isinstance(param[key], (dict, list)):
                substitute_args(param[key], context, only=only)
            # Otherwise, add the element to the result
            elif isinstance(param[key], str):
                param[key] = resolve_args(param[key], context, only=only)
    elif isinstance(param, list):
        for idx, i in enumerate(param):
            if isinstance(i, (dict, list)):
                substitute_args(i, context, only=only)
            elif isinstance(i, str):
                param[idx] = resolve_args(i, context, only=only)
    return param


def resolve_args(arg_str, context=None, resolve_anon=True, filename=None, only=None):
    """
    Resolves substitution args (see wiki spec U{http://ros.org/wiki/roslaunch}).

    @param arg_str: string to resolve zero or more substitution args
        in. arg_str may be None, in which case resolve_args will
        return None
    @type  arg_str: str
    @param context dict: (optional) dictionary for storing results of
        the 'anon' and 'arg' substitution args. multiple calls to
        resolve_args should use the same context so that 'anon'
        substitions resolve consistently. If no context is provided, a
        new one will be created for each call. Values for the 'arg'
        context should be stored as a dictionary in the 'arg' key.
    @type  context: dict
    @param resolve_anon bool: If True (default), will resolve $(anon
        foo). If false, will leave these args as-is.
    @type  resolve_anon: bool

    @return str: arg_str with substitution args resolved
    @rtype:  str
    @raise roslaunch.substitution_args.SubstitutionException: if there is an error resolving substitution args
    """
    if context is None:
        context = {}
    if not arg_str:
        return arg_str
    # special handling of $(eval ...)
    if arg_str.startswith("$(eval ") and arg_str.endswith(")"):
        return roslaunch.substitution_args._eval(arg_str[7:-1], context)
    # first resolve variables like 'env' and 'arg'
    commands = {
        "env": roslaunch.substitution_args._env,
        "optenv": roslaunch.substitution_args._optenv,
        "dirname": roslaunch.substitution_args._dirname,
        "anon": roslaunch.substitution_args._anon,
        "arg": roslaunch.substitution_args._arg,
        "ns": _ns,
        "config": _config,
    }
    if only is not None:
        exec_commands = {}
        for c in only:
            if c in commands:
                exec_commands[c] = commands[c]
    else:
        exec_commands = commands
    resolved = _resolve_args(arg_str, context, resolve_anon, exec_commands)
    # then resolve 'find' as it requires the subsequent path to be expanded already
    commands = {
        "find": roslaunch.substitution_args._find,
    }
    if only is not None:
        exec_commands = {}
        for c in only:
            if c in commands:
                exec_commands[c] = commands[c]
    else:
        exec_commands = commands
    resolved = _resolve_args(resolved, context, resolve_anon, exec_commands)
    return resolved


def _resolve_args(arg_str, context, resolve_anon, commands):
    ros_valid = ["find", "env", "optenv", "dirname", "anon", "arg"]
    valid = ros_valid + ["ns", "config"]
    resolved = arg_str
    if isinstance(arg_str, (str, list)):
        for a in roslaunch.substitution_args._collect_args(arg_str):
            splits = [s for s in a.split(" ") if s]
            if not splits[0] in valid:
                raise roslaunch.substitution_args.SubstitutionException(
                    "Unknown substitution command [%s]. Valid commands are %s" % (a, valid)
                )
            command = splits[0]
            args = splits[1:]
            if command in commands:
                resolved = commands[command](resolved, a, args, context)
    return resolved


def _eval_config(name, args):
    try:
        return args[name]
    except KeyError:
        raise roslaunch.substitution_args.ArgException(name)


_eval_ns = _eval_config


def tryeval(val):
    try:
        val = ast.literal_eval(val)
    except (ValueError, SyntaxError):
        pass
    return val


def _config(resolved, a, args, context):
    """
    process $(config) arg

    :returns: updated resolved argument, ``str``
    :raises: :exc:`roslaunch.substitution_args.ArgException` If arg invalidly specified
    """
    if len(args) == 0:
        raise roslaunch.substitution_args.SubstitutionException("$(config var) must specify a variable name [%s]" % (a))
    elif len(args) > 1:
        raise roslaunch.substitution_args.SubstitutionException("$(config var) may only specify one arg [%s]" % (a))

    if "config" not in context:
        context["config"] = {}
    try:
        return tryeval(resolved.replace("$(%s)" % a, str(_eval_config(name=args[0], args=context["config"]))))
    except roslaunch.substitution_args.ArgException as e:  # roslaunch.substitution_args.ArgException:
        if isinstance(e, roslaunch.substitution_args.ArgException):
            return resolved
        elif isinstance(e, TypeError):
            raise
        else:
            raise


def _ns(resolved, a, args, context):
    """
    process $(ns) arg

    :returns: updated resolved argument, ``str``
    :raises: :exc:`roslaunch.substitution_args.ArgException` If arg invalidly specified
    """
    if len(args) == 0:
        raise roslaunch.substitution_args.SubstitutionException("$(ns var) must specify a variable name [%s]" % (a))
    elif len(args) > 1:
        raise roslaunch.substitution_args.SubstitutionException("$(ns var) may only specify one arg [%s]" % (a))

    if "ns" not in context:
        context["ns"] = {}
    try:
        return tryeval(resolved.replace("$(%s)" % a, str(_eval_ns(name=args[0], args=context["ns"]))))
    except roslaunch.substitution_args.ArgException as e:  # roslaunch.substitution_args.ArgException:
        if isinstance(e, roslaunch.substitution_args.ArgException):
            return resolved
        elif isinstance(e, TypeError):
            raise
        else:
            raise


Stamp = NamedTuple("Stamp", [("seq", int), ("sim_stamp", float), ("wc_stamp", float)])
Stamp.__new__.__defaults__ = (None,) * len(Stamp._fields)
Info = NamedTuple(
    "Info",
    [
        ("name", str),
        ("node_tick", int),
        ("rate_in", float),
        ("t_node", List[Stamp]),
        ("t_in", List[Stamp]),
        ("done", bool),
    ],
)
Info.__new__.__defaults__ = (None,) * len(Info._fields)
Msg = NamedTuple("Msg", [("info", Info), ("msgs", List[Any])])


def check_valid_rosparam_type(param):
    valid_types = (str, int, list, float, bool, dict)
    if isinstance(param, valid_types) or param is None:
        if isinstance(param, dict):
            for key, value in param.items():
                assert isinstance(
                    key, str
                ), 'Only keys of type "str" are supported in dictionaries that are uploaded to the rosparam server.'
                check_valid_rosparam_type(value)
        if isinstance(param, list):
            for value in param:
                check_valid_rosparam_type(value)
    else:
        raise ValueError(
            'Type "%s" of a specified param with value "%s" is not supported by the rosparam server.'
            % (type(param), param.__name__)
        )


def msg_type_error(
    source,
    target,
    msg_type_out,
    converter_out,
    msg_type_ros,
    converter_in,
    msg_type_in,
    msg_type_in_yaml,
):
    if isinstance(source, tuple):
        source = list(source)
    if isinstance(target, tuple):
        target = list(target)
    msg_type_str = '\n\nConversion of msg_type from source="%s/%s/%s" ---> target="%s/%s/%s":\n\n' % tuple(source + target)
    msg_type_str += "".join(
        (f">> msg_type_source:  {msg_type_out} (as specified in source)\n         ||\n         ", r"\/", "\n")
    )
    msg_type_str += "".join((f">> output_converter: {converter_out} \n         ||\n         ", r"\/", "\n"))
    msg_type_str += "".join((f">> msg_type_ROS:     {msg_type_ros} \n         ||\n         ", r"\/", "\n"))
    msg_type_str += "".join((f">> input_converter:  {converter_in} \n         ||\n         ", r"\/", "\n"))
    msg_type_str += "".join(
        (
            f">> msg_type_target:  {msg_type_in} (inferred from converters)\n         ",
            r"/\ ",
            "\n",
            "         || (These must be equal, but they are not!!)\n         ",
            r"\/",
            "\n",
        )
    )
    msg_type_str += f">> msg_type_target:  {msg_type_in_yaml} (as specified in target)\n"
    return msg_type_str


def deepcopy(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return copy.deepcopy(func(*args, **kwargs))

    return wrapper


def is_supported_type(param: Any, types: Tuple, none_support):
    if isinstance(param, types) or (param is None and none_support):
        if isinstance(param, dict):
            for key, value in param.items():
                assert isinstance(key, str), f'Invalid key "{key}". Only type "str" is supported as dictionary key.'
                is_supported_type(value, types, none_support)
        elif not isinstance(param, str) and hasattr(param, "__iter__"):
            for value in param:
                is_supported_type(value, types, none_support)
    else:
        raise TypeError(
            f'Type "{type(param)}" of a specified (nested) param "{param}" is not supported. Only types {types} are supported.'
        )


def supported_types(*types: Tuple, is_classmethod=True):
    # Check if we support NoneType
    none_support = False
    for a in types:
        if a is None:
            none_support = True
            break

    # Remove None from types
    types = tuple([t for t in types if t is not None])

    def _check(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if is_classmethod:
                check_args = list(args[1:]) + [value for _, value in kwargs.items()]
            else:
                check_args = list(args) + [value for _, value in kwargs.items()]
            for param in check_args:
                is_supported_type(param, types, none_support)
            return func(*args, **kwargs)

        return wrapper

    return _check


def get_default_params(func):
    argspec = inspect.getfullargspec(func)
    if argspec.defaults:
        positional_count = len(argspec.args) - len(argspec.defaults)
        defaults = dict(zip(argspec.args[positional_count:], argspec.defaults))
    else:
        defaults = dict()
        positional_count = len(argspec.args)
    for arg in argspec.args[:positional_count]:
        if arg == "self":
            continue
        defaults[arg] = None
    return defaults



#!/usr/bin/env python3

import rospy
from std_msgs.msg import UInt64

# Rx imports
import eagerx.core.rx_message_broker
import eagerx.core.rx_operators
import eagerx.core.rx_pipelines
from eagerx.utils.utils import (
    get_attribute_from_module,
    initialize_converter,
    get_param_with_blocking,
    get_opposite_msg_cls,
)
from eagerx.core.executable_node import RxNode
from eagerx.utils.node_utils import wait_for_node_initialization
from eagerx.core.constants import log_levels_ROS

# Other imports
from threading import Condition
import sys


class RxBridge(object):
    def __init__(self, name, message_broker):
        self.name = name
        self.ns = "/".join(name.split("/")[:2])
        self.mb = message_broker
        self.initialized = False
        self.has_shutdown = False

        # Prepare input & output topics
        (
            rate,
            inputs,
            outputs,
            states,
            node_names,
            target_addresses,
            self.bridge,
        ) = self._prepare_io_topics(self.name)

        # Initialize reactive pipeline
        rx_objects = eagerx.core.rx_pipelines.init_bridge(
            self.ns,
            rate,
            self.bridge,
            inputs,
            outputs,
            states,
            node_names,
            target_addresses,
            self.mb,
        )
        self.mb.add_rx_objects(node_name=name, node=self, **rx_objects)
        self.mb.connect_io()
        self.cond_reg = Condition()

        # Prepare closing routine
        rospy.on_shutdown(self.node_shutdown)

    def node_initialized(self):
        with self.cond_reg:
            # Wait for all nodes to be initialized
            wait_for_node_initialization(self.bridge.is_initialized)

            # Notify env that node is initialized
            if not self.initialized:
                self.init_pub = rospy.Publisher(self.name + "/initialized", UInt64, queue_size=0, latch=True)
                self.init_pub.publish(UInt64(data=1))
                rospy.loginfo('Node "%s" initialized.' % self.name)
                self.initialized = True

    def _prepare_io_topics(self, name):
        params = get_param_with_blocking(name)
        node_names = params["node_names"]
        target_addresses = params["target_addresses"]
        rate = params["rate"]

        # Get node
        node_cls = get_attribute_from_module(params["node_type"])
        node = node_cls(ns=self.ns, message_broker=self.mb, **params)

        # Prepare input topics
        for i in params["inputs"]:
            i["msg_type"] = get_attribute_from_module(i["msg_type"])
            i["msg_type"] = get_opposite_msg_cls(i["msg_type"], i["converter"])
            if isinstance(i["converter"], dict):
                i["converter"] = initialize_converter(i["converter"])
            # else:  # Converter already initialized

        # Prepare output topics
        for i in params["outputs"]:
            i["msg_type"] = get_attribute_from_module(i["msg_type"])
            i["msg_type"] = get_opposite_msg_cls(i["msg_type"], i["converter"])
            if isinstance(i["converter"], dict):
                i["converter"] = initialize_converter(i["converter"])
            # else:  # Converter already initialized

        # Prepare state topics
        for i in params["states"]:
            i["msg_type"] = get_attribute_from_module(i["msg_type"])
            i["msg_type"] = get_opposite_msg_cls(i["msg_type"], i["converter"])
            if isinstance(i["converter"], dict):
                i["converter"] = initialize_converter(i["converter"])
            # else:  # Converter already initialized

        return (
            rate,
            params["inputs"],
            tuple(params["outputs"]),
            tuple(params["states"]),
            node_names,
            target_addresses,
            node,
        )

    def _shutdown(self):
        rospy.logdebug(f"[{self.name}] RxBridge._shutdown() called.")
        self.init_pub.unregister()

    def node_shutdown(self):
        if not self.has_shutdown:
            rospy.logdebug(f"[{self.name}] RxBridge.node_shutdown() called.")
            for address, node in self.bridge.launch_nodes.items():
                rospy.loginfo(f"[{self.name}] Send termination signal to '{address}'.")
                node.terminate()
            for _, rxnode in self.bridge.sp_nodes.items():
                rxnode: RxNode
                if not rxnode.has_shutdown:
                    rospy.loginfo(f"[{self.name}] Shutting down '{rxnode.name}'.")
                    rxnode.node_shutdown()
            rospy.loginfo(f"[{self.name}] Shutting down.")
            self._shutdown()
            self.bridge.shutdown()
            self.mb.shutdown()
            self.has_shutdown = True


if __name__ == "__main__":
    try:
        executable, ns, name = sys.argv[0], sys.argv[-2], sys.argv[-1]

        log_level = get_param_with_blocking(ns + "/log_level")

        rospy.init_node(
            f"{name}".replace("/", "_"),
            log_level=log_levels_ROS[log_level],
            anonymous=True,
        )

        message_broker = eagerx.core.rx_message_broker.RxMessageBroker(owner=f"{ns}/{name}")

        pnode = RxBridge(name=f"{ns}/{name}", message_broker=message_broker)

        message_broker.connect_io()

        pnode.node_initialized()

        rospy.spin()
    finally:
        if not pnode.has_shutdown:
            rospy.loginfo(f"[{ns}/{name}] Send termination signal to '{ns}/{name}'.")
            rospy.signal_shutdown(f"[{ns}/{name}] Terminating '{ns}/{name}'.")



# ROS IMPORTS
from sensor_msgs.msg import Image
from std_msgs.msg import Float32MultiArray, Float32, Bool

# RX IMPORTS
from eagerx.core import register as register
from eagerx.core.entities import SpaceConverter
from eagerx.core.specs import ConverterSpec

# OTHER
import numpy as np
from gym.spaces import Box


class Space_Float32MultiArray(SpaceConverter):
    MSG_TYPE_A = np.ndarray
    MSG_TYPE_B = Float32MultiArray

    @staticmethod
    @register.spec("Space_Float32MultiArray", SpaceConverter)
    def spec(spec: ConverterSpec, low=None, high=None, dtype="float32"):
        # Initialize spec with default arguments
        spec.initialize(Space_Float32MultiArray)
        params = dict(low=low, high=high, dtype=dtype)
        spec.config.update(params)

    def initialize(self, low=None, high=None, dtype="float32"):
        self.low = np.array(low, dtype=dtype)
        self.high = np.array(high, dtype=dtype)
        self.dtype = dtype

    def get_space(self):
        # return Box(self.low, self.high, shape=[1,], dtype=self.dtype)
        return Box(self.low, self.high, dtype=self.dtype)

    def A_to_B(self, msg):
        return Float32MultiArray(data=msg)

    def B_to_A(self, msg):
        return np.array(msg.data, dtype=self.dtype)


class Space_Image(SpaceConverter):
    MSG_TYPE_A = np.ndarray
    MSG_TYPE_B = Image

    @staticmethod
    @register.spec("Space_Image", SpaceConverter)
    def spec(spec: ConverterSpec, low=None, high=None, shape=None, dtype="uint64"):
        # Initialize spec with default arguments
        spec.initialize(Space_Image)

        params = dict(low=low, high=high, shape=shape, dtype=dtype)
        spec.config.update(params)

    def initialize(self, low=None, high=None, shape=None, dtype="float32"):
        if shape is not None and len(shape) == 2:
            shape = shape.append(3)
        self.low = low
        self.high = high
        if isinstance(low, list):
            self.low = np.array(self.low)
        if isinstance(high, list):
            self.high = np.array(self.high)
        if shape is not None:
            assert (
                isinstance(low, (int, float)) or self.low.shape == shape
            ), "If a shape is defined, low must be of a list or type (int, float)."
            assert (
                isinstance(high, (int, float)) or self.high.shape == shape
            ), "If a shape is defined, high must be a list of type (int, float)."
            self.low = low
            self.high = high
        self.shape = shape
        self.dtype = dtype

    def get_space(self):
        return Box(self.low, self.high, shape=self.shape, dtype=self.dtype)

    def A_to_B(self, msg):
        image = self._change_dtype_np_image(msg, "uint8")
        height = image.shape[0]
        width = image.shape[1]
        msg = Image(data=image.reshape(-1).tolist(), height=height, width=width)
        return msg

    def B_to_A(self, msg):
        image = np.frombuffer(msg.data, dtype=np.uint8).reshape(msg.height, msg.width, -1)
        self._change_dtype_np_image(image, self.dtype)
        return image

    def _change_dtype_np_image(self, image, dtype):
        if dtype and image.dtype != dtype:
            if image.dtype in ("float32", "float64") and dtype == "uint8":
                image = (image * 255).astype(dtype)
            elif image.dtype == "uint8" and dtype in ("float32", "float64"):
                image = image.astype(dtype) / 255
            else:
                message = "Cannot convert observations from {} to {}."
                raise NotImplementedError(message.format(image.dtype, dtype))
        return image


class Space_Float32(SpaceConverter):
    MSG_TYPE_A = np.ndarray
    MSG_TYPE_B = Float32

    @staticmethod
    @register.spec("Space_Float32", SpaceConverter)
    def spec(spec: ConverterSpec, low=None, high=None, dtype="float32"):
        # Initialize spec with default arguments
        spec.initialize(Space_Float32)

        params = dict(low=low, high=high, dtype=dtype)
        spec.config.update(params)

    def initialize(self, low=None, high=None, dtype="float32"):
        self.low = low
        self.high = high
        assert isinstance(low, (int, float)), "Low must be of type (int, float)."
        assert isinstance(high, (int, float)), "High must be of type (int, float)."
        self.dtype = dtype

    def get_space(self):
        return Box(
            self.low,
            self.high,
            shape=[
                1,
            ],
            dtype=self.dtype,
        )

    def A_to_B(self, msg):
        return Float32(data=msg)

    def B_to_A(self, msg):
        return msg.data


class Space_Bool(SpaceConverter):
    MSG_TYPE_A = np.ndarray
    MSG_TYPE_B = Bool

    @staticmethod
    @register.spec("Space_Bool", SpaceConverter)
    def spec(spec: ConverterSpec):
        pass

    def initialize(self):
        self.dtype = np.bool

    def get_space(self):
        return Box(
            False,
            True,
            shape=[
                1,
            ],
            dtype=self.dtype,
        )

    def A_to_B(self, msg):
        return Float32(data=msg)

    def B_to_A(self, msg):
        return msg.data



# OTHER IMPORTS
from typing import Optional, List

# ROS IMPORTS
from std_msgs.msg import Float32MultiArray, Bool, Float32
from sensor_msgs.msg import Image

# EAGERx IMPORTS
from eagerx.bridges.openai_gym.bridge import GymBridge
from eagerx.core.entities import Object, EngineNode, SpaceConverter
from eagerx.core.specs import ObjectSpec
from eagerx.core.graph_engine import EngineGraph
import eagerx.core.register as register


class GymObject(Object):
    entity_id = "GymObject"

    @staticmethod
    @register.sensors(observation=Float32MultiArray, reward=Float32, done=Bool, image=Image)
    @register.actuators(action=Float32MultiArray)
    @register.simstates()
    @register.config(env_id=None, always_render=False, default_action=None, render_shape=[200, 200])
    def agnostic(spec: ObjectSpec, rate):
        """Agnostic definition of the GymObject"""
        # Register standard converters, space_converters, and processors
        import eagerx.converters  # noqa # pylint: disable=unused-import

        # Set observation space_converters
        spec.sensors.observation.space_converter = SpaceConverter.make(
            "GymSpace_Float32MultiArray", gym_id=spec.config.env_id, space="observation"
        )
        spec.sensors.reward.space_converter = SpaceConverter.make("Space_Float32", low=-99999, high=9999, dtype="float32")
        spec.sensors.done.space_converter = SpaceConverter.make("Space_Bool")
        spec.sensors.image.space_converter = SpaceConverter.make(
            "Space_Image", low=0, high=1, shape=spec.config.render_shape, dtype="float32"
        )
        spec.actuators.action.space_converter = SpaceConverter.make(
            "GymSpace_Float32MultiArray", gym_id=spec.config.env_id, space="action"
        )

        # Set observation rates
        spec.sensors.observation.rate = rate
        spec.sensors.reward.rate = rate
        spec.sensors.done.rate = rate
        spec.sensors.image.rate = rate
        spec.actuators.action.rate = rate

    @staticmethod
    @register.spec(entity_id, Object)
    def spec(
        spec: ObjectSpec,
        name: str,
        sensors: Optional[List[str]] = None,
        env_id: str = "Pendulum-v1",
        rate: float = 20.0,
        always_render: bool = False,
        default_action=None,
        render_shape: Optional[List[int]] = None,
    ):
        """Object spec of GymObject"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        GymObject.initialize_spec(spec)

        # Set default node params
        spec.config.name = name
        spec.config.sensors = sensors if sensors else ["observation", "reward", "done"]
        spec.config.actuators = ["action"]

        # Set custom node params
        spec.config.render_shape = render_shape if render_shape else [200, 200]
        spec.config.env_id = env_id
        spec.config.always_render = always_render

        # Add agnostic definition
        GymObject.agnostic(spec, rate)

    @staticmethod
    @register.bridge(entity_id, GymBridge)  # This decorator pre-initializes bridge implementation with default object_params
    def openai_gym(spec: ObjectSpec, graph: EngineGraph):
        """Engine-specific implementation (GymBridge) of the object."""
        # Set bridge arguments (nothing to set here in this case)
        spec.GymBridge.env_id = spec.config.env_id

        # Create sensor engine nodes
        # Rate=None, because we will connect them to sensors (thus uses the rate set in the agnostic specification)
        obs = EngineNode.make("ObservationSensor", "obs", rate=spec.sensors.observation.rate, process=2)
        rwd = EngineNode.make("RewardSensor", "rwd", rate=spec.sensors.reward.rate, process=2)
        done = EngineNode.make("DoneSensor", "done", rate=spec.sensors.done.rate, process=2)
        image = EngineNode.make(
            "GymImage",
            "image",
            shape=spec.config.render_shape,
            always_render=spec.config.always_render,
            rate=spec.sensors.image.rate,
            process=2,
        )

        # Create actuator engine nodes
        # Rate=None, because we will connect it to an actuator (thus uses the rate set in the agnostic specification)
        action = EngineNode.make(
            "ActionActuator", "action", rate=spec.actuators.action.rate, process=2, zero_action=spec.config.default_action
        )

        # Connect all engine nodes
        graph.add([obs, rwd, done, image, action])
        graph.connect(source=obs.outputs.observation, sensor="observation")
        graph.connect(source=rwd.outputs.reward, sensor="reward")
        graph.connect(source=done.outputs.done, sensor="done")
        graph.connect(source=image.outputs.image, sensor="image")
        graph.connect(actuator="action", target=action.inputs.action)

        # Check graph validity (comment out)
        # graph.is_valid(plot=True)



import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.lines import Line2D
import networkx as nx


def reset_graph(G):
    # Create a shallow copy graph that excludes feedthrough edges
    F = nx.MultiDiGraph(G)
    for u, v, key, data in G.edges(data=True, keys=True):
        if data["feedthrough"]:
            F.remove_edge(u, v, key=key)

    # Set actions node to stale
    if "observations" in F.nodes:
        F.nodes["observations"]["is_stale"] = True
    elif "env/observations" in F.nodes:
        F.nodes["env/observations"]["is_stale"] = True
    else:
        raise KeyError("Observation node not in graph.")
    return F


def episode_graph(G):
    # Create a shallow copy graph that excludes "skip" edges
    H = nx.MultiDiGraph(G)
    for u, v, key, data in G.edges(data=True, keys=True):
        if data["skip"]:
            H.remove_edge(u, v, key=key)

    # Color cyclic edges red
    cycles = []
    C = list(nx.simple_cycles(H))
    for c in C:
        H_sub = H.subgraph(c)
        for u, v, key, data in H_sub.edges(data=True, keys=True):
            G[u][v][key]["color"] = "red"
            data["color"] = "red"

        # Get all simple cyclic edge paths
        out_nodes = []
        for _u, v, _key, data in H_sub.out_edges(c[0], data=True, keys=True):
            out_nodes.append((v, data))

        for n, data_first in out_nodes:
            paths = nx.all_simple_edge_paths(H_sub, n, c[0])
            first_edge = (data_first["source"], data_first["target"])
            for path in paths:
                edges = [first_edge]
                for edge in path:
                    data = H_sub.get_edge_data(*edge)
                    edge = (data["source"], data["target"])
                    edges.append(edge)
                cycles.append(edges)
    return H, cycles


def plot_graph(G, ax=None, k=2, pos=None):
    if ax is None:
        env, ax = plt.subplots(nrows=1, ncols=1)

    # Determine position
    if pos is None:
        pos = nx.spring_layout(G, k=k)

    # Plot labels
    bbox = dict(facecolor="skyblue", edgecolor="black", boxstyle="round,pad=0.2", alpha=0.5)
    labels = nx.draw_networkx_labels(G, pos, bbox=bbox, ax=ax)
    for n, data in G.nodes(data=True):
        facecolor = data["facecolor"]
        labels[n].get_bbox_patch().set_facecolor(facecolor)

    # Plot nodes
    nodes_plt = nx.draw_networkx_nodes(G, pos, node_color="white", node_size=1500, alpha=0, ax=ax)

    # Plot edges
    colors = [data["color"] for u, v, data in G.edges(data=True)]
    styles = [data["style"] for u, v, data in G.edges(data=True)]
    alphas = [data["alpha"] for u, v, data in G.edges(data=True)]
    lines = nx.draw_networkx_edges(G, pos, arrows=True, edge_color=colors, width=2, ax=ax)
    for style, alpha, line in zip(styles, alphas, lines):
        line.set_linestyle(style)
        line.set_alpha(alpha)

    # Plot nodes
    nodes_plt = nx.draw_networkx_nodes(G, pos, node_color="white", node_size=350, alpha=1.0, ax=ax)

    # Create legend
    root_patch = mpatches.Patch(color="skyblue", label="source")
    sink_patch = mpatches.Patch(color="lightgreen", label="sink")
    stale_patch = mpatches.Patch(color="lightgrey", label="stale")
    node_patch = mpatches.Patch(color="khaki", label="node")
    stale_line = Line2D([0], [0], label="stale", color="lightgrey")
    cyclic_line = Line2D([0], [0], label="cyclic", color="red")
    dag_line = Line2D([0], [0], label="acyclic", color="black")
    nondag_line = Line2D([0], [0], label="skipped", color="green", linestyle="dotted")
    ax.legend(
        handles=[
            root_patch,
            sink_patch,
            node_patch,
            stale_patch,
            dag_line,
            cyclic_line,
            nondag_line,
            stale_line,
        ],
        ncol=4,
        prop={"size": 8},
        loc="upper center",
        bbox_to_anchor=(0.5, -0.05),
        fancybox=True,
        shadow=True,
    )
    return nodes_plt, lines, labels, pos


def color_nodes(G):
    # Color nodes based on in/out going edges
    for n, data in G.nodes(data=True):
        if data["is_stale"]:
            facecolor = "lightgrey"
        else:
            in_edges = G.in_degree(n)
            out_edges = G.out_degree(n)
            if in_edges == 0 and out_edges == 0:
                facecolor = "lightgrey"
            elif in_edges > 0 and out_edges == 0:
                facecolor = "lightgreen"
            elif in_edges == 0 and out_edges > 0:
                facecolor = "skyblue"
            else:  # in_edges > 0 and out_edges > 0:
                facecolor = "khaki"
        data["facecolor"] = facecolor


def color_edges(G):
    for _u, _v, _key, data in G.edges(data=True, keys=True):
        if data["is_stale"]:
            data["alpha"] = 0.3
        else:
            data["alpha"] = 1.0


def is_stale(G, exclude_skip=False):
    new_stale_nodes = [n for n, is_stale in nx.get_node_attributes(G, "is_stale").items() if is_stale]
    for n, data_n in G.nodes(data=True):
        # First, determine stale nodes, based on edges_in & always connected
        in_degree = G.in_degree(n)
        if in_degree == 0 and not data_n["always_active"] and not data_n["has_tick"]:
            data_n["is_stale"] = True
            new_stale_nodes.append(n)

        # Then, set all edges to active (=not stale)
        for _u, _v, _key, data_e in G.edges(n, data=True, keys=True):
            data_e["is_stale"] = False

    # Now, iteratively run over all new stale nodes and set out_edges and out_nodes to stale.
    # We add out_nodes (previously not stale) to new_stale nodes
    while len(new_stale_nodes) > 0:
        stale_nodes = new_stale_nodes
        new_stale_nodes = []

        # Then, loop over new stale nodes of prev. iteration and set all out-going edges to stale
        for n in stale_nodes:
            for u, v, _key, data_e in G.out_edges(n, data=True, keys=True):
                data_e["is_stale"] = True
                skip = data_e["skip"] if exclude_skip else False
                if u == "N8/out_1" and v == "N7/out_1":
                    print("wait")
                if not G.nodes[v]["is_stale"] and not skip:
                    if not G.nodes[v]["always_active"]:
                        new_stale_nodes.append(v)
                        G.nodes[v]["is_stale"] = True
    remain_active = [n for n, is_stale in nx.get_node_attributes(G, "remain_active").items() if is_stale]
    stale_nodes = [n for n, is_stale in nx.get_node_attributes(G, "is_stale").items() if is_stale]
    not_active = [n for n in stale_nodes if n in remain_active]
    return not_active



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eagerx',
 'eagerx.bridges',
 'eagerx.bridges.openai_gym',
 'eagerx.bridges.test',
 'eagerx.converters',
 'eagerx.core',
 'eagerx.enginestates',
 'eagerx.nodes',
 'eagerx.utils',
 'eagerx.wrappers']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'Rx>=3.2.0,<4.0.0',
 'Sphinx>=4.0,<5.0',
 'defusedxml>=0.7.1,<0.8.0',
 'gym>=0.20.0,<0.21.0',
 'netifaces>=0.11.0,<0.12.0',
 'networkx>=2.5.1,<3.0.0',
 'opencv-python>=4.3.0.36,<5.0.0.0',
 'psutil>=5.9.0,<6.0.0',
 'pyglet>=1.5.21,<2.0.0',
 'rospkg>=1.3.0,<2.0.0',
 'scikit-image>=0.17.2,<0.18.0',
 'sphinx-autodoc-typehints>=1.0,<2.0',
 'sphinx-rtd-theme>=1.0,<2.0',
 'tabulate>=0.8.9,<0.9.0',
 'termcolor>=1.1.0,<2.0.0',
 'typing>=3.0,<4.0']

setup_kwargs = {
    'name': 'eagerx',
    'version': '0.1.9',
    'description': 'Engine Angostic Gym Environments for Robotics',
    'long_description': None,
    'author': 'Bas van der Heijden',
    'author_email': 'd.s.vanderheijden@tudelft.nl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/eager-dev/eagerx',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



import gym
from gym.core import ActionWrapper
from gym.wrappers import FlattenObservation


class FlattenAction(ActionWrapper):
    """
    Observation wrapper that flattens the observation.

    :param env: The environment to flatten
    """

    def __init__(self, env: gym.Env) -> None:
        super(FlattenAction, self).__init__(env)
        self.action_space = gym.spaces.flatten_space(env.action_space)

    def action(self, action: object) -> object:
        return gym.spaces.unflatten(self.env.action_space, action)


class Flatten(gym.Wrapper):
    """
    Flattens the environment action and observations space to a single
    type.

    :param env: The environment to flatten
    """

    def __init__(self, env: gym.Env) -> None:
        super().__init__(FlattenObservation(FlattenAction(env)))



# ROS packages required
import rospy
import rosparam
from rosgraph.masterapi import ROSMasterException
from std_msgs.msg import UInt64
from rosgraph.masterapi import Error
from sensor_msgs.msg import Image

# EAGERX
from eagerx.core.specs import NodeSpec, ObjectSpec, BridgeSpec
from eagerx.core.entities import Node
from eagerx.core.graph import Graph
from eagerx.utils.node_utils import (
    initialize_nodes,
    wait_for_node_initialization,
    substitute_args,
)
from eagerx.core.executable_node import RxNode
from eagerx.core.executable_bridge import RxBridge
from eagerx.core.supervisor import Supervisor, SupervisorNode
from eagerx.core.rx_message_broker import RxMessageBroker
from eagerx.core.constants import process

# OTHER IMPORTS
import atexit
import abc
import cv2
import numpy as np
from copy import deepcopy
from typing import List, Union, Dict, Tuple, Callable
import gym
import logging


class Env(gym.Env):
    @staticmethod
    def create_supervisor():
        entity_type = f"{SupervisorNode.__module__}/{SupervisorNode.__name__}"
        supervisor = Node.pre_make("N/a", entity_type)
        supervisor.add_output("step", msg_type=UInt64)

        supervisor.config.name = "env/supervisor"
        supervisor.config.color = "yellow"
        supervisor.config.process = process.ENVIRONMENT
        supervisor.config.outputs = ["step"]
        return supervisor

    def __init__(self, name: str, rate: float, graph: Graph, bridge: BridgeSpec) -> None:
        assert "/" not in name, 'Environment name "%s" cannot contain the reserved character "/".' % name
        self.name = name
        self.ns = "/" + name
        self.rate = rate
        self.initialized = False
        self.has_shutdown = False

        # Take deepcopy of bridge
        bridge = BridgeSpec(bridge.params)
        self._bridge_name = bridge.params["config"]["entity_id"]

        # Register graph
        self.graph = graph
        nodes, objects, actions, observations, self.render_node = graph.register()

        # Add bridge implementation
        [o.add_bridge(self._bridge_name) for o in objects]

        # Initialize supervisor node
        self.mb, self.supervisor_node, self.supervisor = self._init_supervisor(bridge, nodes, objects)
        self._is_initialized = self.supervisor_node.is_initialized

        # Initialize bridge
        self._init_bridge(bridge, nodes)

        # Create environment node
        self.env_node, self.env = self._init_environment(actions, observations, self.supervisor_node, self.mb)

        # Register render node
        if self.render_node:
            nodes = [self.render_node] + nodes

        # Register nodes
        self.register_nodes(nodes)

        # Register objects
        self.register_objects(objects)

        # Implement clean up
        atexit.register(self.shutdown)

    def _init_supervisor(self, bridge: BridgeSpec, nodes: List[NodeSpec], objects: List[ObjectSpec]):
        # Initialize supervisor
        supervisor = self.create_supervisor()

        # Get all states from objects & nodes
        for i in [bridge] + nodes + objects:
            if "states" not in i.params["config"]:
                continue
            for cname in i.params["config"]["states"]:
                entity_name = i.config.name
                name = f"{entity_name}/{cname}"
                address = f"{entity_name}/states/{cname}"
                msg_type = i.params["states"][cname]["msg_type"]
                space_converter = i.params["states"][cname]["space_converter"]

                assert (
                    name not in supervisor.params["states"]
                ), f'Cannot have duplicate states. State "{name}" is defined multiple times.'

                mapping = dict(address=address, msg_type=msg_type, converter=space_converter)
                with supervisor.states as d:
                    d[name] = mapping
                supervisor.config.states.append(name)

            # Get states from simnodes. WARNING: can make environment non-agnostic.
            if isinstance(i, ObjectSpec):
                obj_name = i.config.name
                context = {"ns": {"obj_name": obj_name}, "config": i.config.to_dict()}
                for node_name, params_simnode in i.params[self._bridge_name]["nodes"].items():
                    if "states" in params_simnode["config"]:
                        for cname in params_simnode["config"]["states"]:
                            comp_params = params_simnode["states"][cname]
                            node_name_sub = substitute_args(node_name, context=context, only=["ns", "config"])
                            name = f"{node_name_sub}/{cname}"
                            address = f"{node_name_sub}/states/{cname}"
                            msg_type = comp_params["msg_type"]
                            space_converter = comp_params["space_converter"]

                            rospy.logwarn(
                                f'Adding state "{name}" to simulation node "{node_name_sub}" can potentially make the agnostic environment with object "{entity_name}" engine-specific. Check the spec of "{i.config.entity_id}" under bridge implementation "{self._bridge_name}" for more info.'
                            )
                            assert (
                                name not in supervisor.params["states"]
                            ), f'Cannot have duplicate states. State "{name}" is defined multiple times.'

                            mapping = dict(
                                address=address,
                                msg_type=msg_type,
                                converter=space_converter,
                            )
                            with supervisor.states as d:
                                d[name] = mapping
                            supervisor.config.states.append(name)

        # Delete pre-existing parameters
        try:
            rosparam.delete_param(f"/{self.name}")
            rospy.loginfo(f'Pre-existing parameters under namespace "/{self.name}" deleted.')
        except Error:
            pass

        # Upload log_level
        log_level = logging.getLogger("rosout").getEffectiveLevel()
        rosparam.upload_params(self.ns, {"log_level": log_level})

        # Initialize message broker
        mb = RxMessageBroker(owner="%s/%s" % (self.ns, "env"))

        # Get info from bridge on reactive properties
        is_reactive = bridge.config.is_reactive
        real_time_factor = bridge.config.real_time_factor
        simulate_delays = bridge.config.simulate_delays

        # Create supervisor node
        name = supervisor.config.name
        supervisor.config.rate = self.rate
        supervisor_params = supervisor.build(ns=self.ns)
        rosparam.upload_params(self.ns, supervisor_params)
        rx_supervisor = Supervisor(
            "%s/%s" % (self.ns, name),
            mb,
            is_reactive,
            real_time_factor,
            simulate_delays,
        )
        rx_supervisor.node_initialized()

        # Connect io
        mb.connect_io()
        return mb, rx_supervisor.node, rx_supervisor

    def _init_bridge(self, bridge: BridgeSpec, nodes: List[NodeSpec]) -> None:
        # Check that reserved keywords are not already defined.
        assert (
            "node_names" not in bridge.params["config"]
        ), f'Keyword "{"node_names"}" is a reserved keyword within the bridge params and cannot be used twice.'
        assert (
            "target_addresses" not in bridge.params["config"]
        ), f'Keyword "{"target_addresses"}" is a reserved keyword within the bridge params and cannot be used twice.'
        assert (
            not bridge.params["config"]["process"] == process.BRIDGE
        ), "Cannot initialize the bridge inside the bridge process, because it has not been launched yet. You can choose process.{ENVIRONMENT, EXTERNAL, NEW_PROCESS}."

        # Extract node_names
        node_names = ["environment", "env/supervisor"]
        target_addresses = []
        for i in nodes:
            # node_names.append(i.params['default']['name'])
            if "targets" in i.params["config"]:
                for cname in i.params["config"]["targets"]:
                    address = i.params["targets"][cname]["address"]
                    target_addresses.append(address)
        with bridge.config as d:
            d.node_names = node_names
            d.target_addresses = target_addresses

        initialize_nodes(
            bridge,
            process.ENVIRONMENT,
            self.ns,
            self.mb,
            self.supervisor_node.is_initialized,
            self.supervisor_node.sp_nodes,
            self.supervisor_node.launch_nodes,
            rxnode_cls=RxBridge,
        )
        wait_for_node_initialization(self._is_initialized)  # Proceed after bridge is initialized

    def _init_environment(self, actions: NodeSpec, observations: NodeSpec, supervisor_node, message_broker):
        # Check that env has at least one input & output.
        assert (
            len(observations.params["config"]["inputs"]) > 0
        ), f'Environment "{self.name}" must have at least one input (i.e. input).'
        assert (
            len(actions.params["config"]["outputs"]) > 0
        ), f'Environment "{self.name}" must have at least one action (i.e. output).'

        # Check that all observation addresses are unique
        addresses_obs = [observations.params["inputs"][cname]["address"] for cname in observations.params["config"]["inputs"]]
        len(set(addresses_obs)) == len(
            addresses_obs
        ), "Duplicate observations found: %s. Make sure to only have unique observations" % (
            set([x for x in addresses_obs if addresses_obs.count(x) > 1])
        )

        # Create env node
        env_spec = Node.make("Environment", rate=self.rate)
        name = env_spec.config.name
        inputs = observations.config.inputs
        outputs = actions.config.outputs
        for i in inputs:
            if i == "actions_set":
                continue
            with env_spec.inputs as d:
                d[i] = getattr(observations.inputs, i)
            env_spec.config.inputs.append(i)
        for i in outputs:
            if i == "set":
                continue
            with env_spec.outputs as d:
                d[i] = getattr(actions.outputs, i)
            env_spec.config.outputs.append(i)
        env_params = env_spec.build(ns=self.ns)
        rosparam.upload_params(self.ns, env_params)
        rx_env = RxNode(name="%s/%s" % (self.ns, name), message_broker=message_broker)
        rx_env.node_initialized()

        # Set env_node in supervisor
        supervisor_node.set_environment(rx_env.node)

        return rx_env.node, rx_env

    @property
    def observation_space(self) -> gym.spaces.Dict:
        assert not self.has_shutdown, "This environment has been shutdown."
        observation_space = dict()
        for name, buffer in self.env_node.observation_buffer.items():
            space = buffer["converter"].get_space()
            if not buffer["window"] > 0:
                continue
            low = np.repeat(space.low[np.newaxis, ...], buffer["window"], axis=0)
            high = np.repeat(space.high[np.newaxis, ...], buffer["window"], axis=0)
            stacked_space = gym.spaces.Box(low=low, high=high, dtype=space.dtype)
            observation_space[name] = stacked_space
        return gym.spaces.Dict(spaces=observation_space)

    @property
    def action_space(self) -> gym.spaces.Dict:
        assert not self.has_shutdown, "This environment has been shutdown."
        action_space = dict()
        for name, buffer in self.env_node.action_buffer.items():
            action_space[name] = buffer["converter"].get_space()
        return gym.spaces.Dict(spaces=action_space)

    @property
    def state_space(self) -> gym.spaces.Dict:
        state_space = dict()
        for name, buffer in self.supervisor_node.state_buffer.items():
            state_space[name] = buffer["converter"].get_space()
        return gym.spaces.Dict(spaces=state_space)

    def _set_action(self, action) -> None:
        # Set actions in buffer
        for name, buffer in self.env_node.action_buffer.items():
            assert not self.supervisor_node.is_reactive or name in action, (
                'Action "%s" not specified. Must specify all actions in action_space if running reactive.' % name
            )
            if name in action:
                buffer["msg"] = action[name]

    def _set_state(self, state) -> None:
        # Set states in buffer
        for name, msg in state.items():
            assert name in self.supervisor_node.state_buffer, 'Cannot set unknown state "%s".' % name
            self.supervisor_node.state_buffer[name]["msg"] = msg

    def _get_observation(self) -> Dict:
        # Get observations from buffer
        observation = dict()
        for name, buffer in self.env_node.observation_buffer.items():
            observation[name] = buffer["msgs"]
        return observation

    def _initialize(self, states: Dict) -> None:
        assert not self.initialized, "Environment already initialized. Cannot re-initialize pipelines. "

        # Set desired reset states
        self._set_state(states)

        # Wait for nodes to be initialized
        [node.node_initialized() for name, node in self.supervisor_node.sp_nodes.items()]
        wait_for_node_initialization(self._is_initialized)

        # Initialize single process communication
        self.mb.connect_io(print_status=True)

        rospy.loginfo("Nodes initialized.")

        # Perform first reset
        self.supervisor_node.reset()

        # Nodes initialized
        self.initialized = True
        rospy.loginfo("Pipelines initialized.")

    def _reset(self, states: Dict) -> Dict:
        assert not self.has_shutdown, "This environment has been shutdown."
        # Initialize environment
        if not self.initialized:
            self._initialize(states)

        # Set desired reset states
        self._set_state(states)

        # Perform reset
        self.supervisor_node.reset()
        obs = self._get_observation()
        return obs

    def _step(self, action: Dict) -> Dict:
        # Check that nodes were previously initialized.
        assert self.initialized, "Not yet initialized. Call .reset() before calling .step()."
        assert not self.has_shutdown, "This environment has been shutdown."

        # Set actions in buffer
        self._set_action(action)

        # Call step
        self.supervisor_node.step()
        return self._get_observation()

    def _shutdown(self):
        if not self.has_shutdown:
            for address, node in self.supervisor_node.launch_nodes.items():
                rospy.loginfo(f"[{self.name}] Send termination signal to '{address}'.")
                node.terminate()
                # node.terminate(f"[{self.name}] Terminating '{address}'")
            for _, rxnode in self.supervisor_node.sp_nodes.items():
                rxnode: RxNode
                if not rxnode.has_shutdown:
                    rospy.loginfo(f"[{self.name}][{rxnode.name}] Shutting down.")
                    rxnode.node_shutdown()
            if not self.supervisor.has_shutdown:
                self.supervisor.node_shutdown()
            if not self.env.has_shutdown:
                self.env.node_shutdown()
            self.mb.shutdown()
            try:
                rosparam.delete_param(f"/{self.name}")
                rospy.loginfo(f'Parameters under namespace "/{self.name}" deleted.')
            except ROSMasterException as e:
                rospy.logwarn(e)
            self.has_shutdown = True

    def register_nodes(self, nodes: Union[List[NodeSpec], NodeSpec]) -> None:
        assert not self.has_shutdown, "This environment has been shutdown."
        # Look-up via <env_name>/<obj_name>/nodes/<component_type>/<component>: /rx/obj/nodes/sensors/pos_sensors
        if not isinstance(nodes, list):
            nodes = [nodes]

        # Register nodes
        [self.supervisor_node.register_node(n) for n in nodes]

    def register_objects(self, objects: Union[List[ObjectSpec], ObjectSpec]) -> None:
        assert not self.has_shutdown, "This environment has been shutdown."
        # Look-up via <env_name>/<obj_name>/nodes/<component_type>/<component>: /rx/obj/nodes/sensors/pos_sensors
        if not isinstance(objects, list):
            objects = [objects]

        # Register objects
        [self.supervisor_node.register_object(o, self._bridge_name) for o in objects]

    def render(self, mode="human"):
        assert not self.has_shutdown, "This environment has been shutdown."
        if self.render_node:
            if mode == "human":
                self.supervisor_node.start_render()
            elif mode == "rgb_array":
                self.supervisor_node.start_render()
                ros_im = self.supervisor_node.get_last_image()
                if ros_im.height == 0 or ros_im.width == 0:
                    # todo: check if channel dim first or last.
                    im = np.empty(shape=(0, 0, 3), dtype=np.uint8)
                else:
                    im = np.frombuffer(ros_im.data, dtype=np.uint8).reshape(ros_im.height, ros_im.width, -1)
                    if "bgr" in ros_im.encoding:
                        # try:
                        # todo: find out what exception to catch here.
                        im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
                        # except :
                        #     pass
                return im
            else:
                raise ValueError('Render mode "%s" not recognized.' % mode)
        else:
            rospy.logwarn_once("No render node active, so not rendering.")
            if mode == "rgb_array":
                return Image()
            else:
                return

    @abc.abstractmethod
    def reset(self) -> Dict:
        pass

    @abc.abstractmethod
    def step(self, action: Dict) -> Tuple[Dict, float, bool, Dict]:
        pass

    def close(self):
        assert not self.has_shutdown, "This environment has been shutdown."
        self.supervisor_node.stop_render()

    def shutdown(self):
        self._shutdown()


class EagerEnv(Env):
    def __init__(
        self,
        name: str,
        rate: float,
        graph: Graph,
        bridge: BridgeSpec,
        step_fn: Callable = lambda prev_obs, obs, action, steps: (obs, 0.0, False, {}),
        reset_fn: Callable = lambda env: env.state_space.sample(),
    ) -> None:
        self.steps = None
        self.prev_observation = None
        self.step_fn = step_fn
        self.reset_fn = reset_fn
        super(EagerEnv, self).__init__(name, rate, graph, bridge)
        self.excl_obs = [name for name, buffer in self.env_node.observation_buffer.items() if buffer["window"] == 0]

    def step(self, action: Dict) -> Tuple[Dict, float, bool, Dict]:
        # Send actions and wait for observations
        observation = self._step(action)
        self.steps += 1

        # Save observation for next step
        prev_obs = deepcopy(observation)

        # Process after step
        observation, reward, is_done, info = self.step_fn(self.prev_observation, observation, action, self.steps)

        # Pop all observations with window = 0 (if present)
        for name in self.excl_obs:
            try:
                observation.pop(name)
            except KeyError:
                pass

        # Update previous observation with current observation (used in next step)
        self.prev_observation = prev_obs
        return observation, reward, is_done, info

    def reset(self) -> Dict:
        # Determine reset states
        states = self.reset_fn(self)

        # Perform reset
        observation = self._reset(states)
        self.prev_observation = deepcopy(observation)

        # Reset number of steps
        self.steps = 0
        return observation



import rospy
from typing import Optional
from eagerx.core.entities import EngineState
import eagerx.core.register as register


class TestEngineState(EngineState):
    @staticmethod
    @register.spec("TestEngineState", EngineState)
    def spec(spec, test_arg: Optional[str] = "test_argument"):
        # Initialize simstate
        spec.initialize(TestEngineState)

        # Modify parameters based on arguments
        spec.config.test_arg = test_arg
        spec.config.test_arg = spec.config.test_arg

    def initialize(self, test_arg):
        self.test_arg = test_arg

    def reset(self, state, done):
        rospy.logdebug("INSIDE SIMSTATE RESET: (%s, %s)" % (state, done))
